@@ -5,6 +5,7 @@ import logging
 from contextlib import suppress
 from typing import Any, Dict, Iterator, List, Tuple, Type
 
+import aiohttp
 from aiohttp import ClientConnectorError, ClientOSError
 from tortoise import Model, fields
 
@@ -148,8 +149,9 @@ async def configure_hasura(config: DipDupConfig):
     _logger.info('Waiting for Hasura instance to be healthy')
     for _ in range(60):
         with suppress(ClientConnectorError, ClientOSError):
-            await http_request('get', url=f'{url}/healthz')
-            break
+            async with aiohttp.ClientSession() as session:
+                await session.get(f'{url}/healthz')
+                break
         await asyncio.sleep(1)
     else:
         _logger.error('Hasura instance not responding for 60 seconds')
